/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

The algorithms found in this code are based on publicly released source from
Scott Scriven.

******************************************************************************

###CLASS###
Name:      Ripple
Version:   1.0
Status:    Stable
Date:      December 2003
ID:        ID_RIPPLE
Category:  Effect
Author:    Rocklyte Systems, Algorithms by Scott Scriven
Copyright: Rocklyte Systems (c) 2002-2003, Scott Scriven (c) 1997.  All rights reserved.
Short:     Creates water ripple effects in real time.
Keywords:  water, puddle, ripple, wave

###DESCRIPTION###
<p>The Ripple class draws an animated effect that emulates the appearance of
water ripples, as well as other wave and vibration based effects.  The
algorithms use a large amount of CPU time and this increases according to the
size of the area that the graphics are being applied to.</p>

<p>Once initialised, the effect will run automatically until the object is
freed from the system.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/input/keyboard.h>
#include <pandora/input/joyport.h>
#include <pandora/graphics/bitmap.h>
#include <pandora/graphics/render.h>
#include <pandora/system/module.h>
#include <pandora/modules/render.h>
#include <math.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_RIPPLE 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_RIPPLE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2002-2003.  All rights reserved.",
   "December 2003",
   "Ripple"
};

static struct KernelBase *KernelBase = 0;
static struct RenderBase *RenderBase;
static OBJECTPTR RenderMod = NULL;
static OBJECTPTR RippleClass = NULL;

/*****************************************************************************
** Ripple class definition.
*/

struct Ripple {
   OBJECT_HEADER
   LONG Flags;
   LONG Strength;
   LONG Density;
   LONG DropHeight;
   LONG CanvasWidth;
   LONG CanvasHeight;

   PRIVATE_FIELDS
   OBJECTID DrawableID;    /* Reference to drawable destination */
   struct Bitmap *Buffer;  /* Internal buffer */
   LONG  Page;              /* Current height page */
   LONG  SwirlAngle;        /* Swirl angle */
   LONG  SurferX;           /* Surfer horizontal offset */
   LONG  SurferY;           /* Surfer vertical offset */
   LONG  SurferXAngle;      /* Surfer horizontal angle */
   LONG  SurferYAngle;      /* Surfer vertical angle */
   LONG  Buttons;           /* Storage for held mouse buttons */
   LONG  Dimensions;
   WORD  *HeightMap[2];
   LONG  DrawableWidth;
   LONG  DrawableHeight;
   FLOAT XCoord;
   FLOAT YCoord;
   FLOAT YOffset;
   FLOAT XOffset;
   FLOAT Width;
   FLOAT Height;
};

#define FLF_SURFER      0x00000001
#define FLF_RAINDROPS   0x00000002
#define FLF_SPLASHES    0x00000004
#define FLF_SWIRL       0x00000008
#define FLF_BLEND       0x00000010
#define FLF_GREYSCALE   0x00000020
#define FLF_PUNCH       0x00000040
#define FLF_WARP        0x00000080
#define FLF_HIGHQUALITY 0x00000100
#define FLF_INTERACTIVE 0x00000200

static struct FieldDef RippleFlags[] = {
   { "SURFER",      FLF_SURFER },
   { "RAINDROPS",   FLF_RAINDROPS },
   { "SPLASHES",    FLF_SPLASHES },
   { "SWIRL",       FLF_SWIRL },
   { "BLEND",       FLF_BLEND },
   { "GREYSCALE",   FLF_GREYSCALE },
   { "PUNCH",       FLF_PUNCH },
   { "WARP",        FLF_WARP },
   { "HIGHQUALITY", FLF_HIGHQUALITY },
   { "INTERACTIVE", FLF_INTERACTIVE },
   { NULL, NULL }
};

static ERROR GET_Height(struct Ripple *, struct Variable *);
static ERROR GET_Width(struct Ripple *, struct Variable *);
static ERROR GET_X(struct Ripple *, struct Variable *);
static ERROR GET_XOffset(struct Ripple *, struct Variable *);
static ERROR GET_Y(struct Ripple *, struct Variable *);
static ERROR GET_YOffset(struct Ripple *, struct Variable *);

static ERROR SET_Density(struct Ripple *, LONG *);
static ERROR SET_Strength(struct Ripple *, LONG *);
static ERROR SET_DropHeight(struct Ripple *, LONG *);
static ERROR SET_Height(struct Ripple *, struct Variable *);
static ERROR SET_Width(struct Ripple *, struct Variable *);
static ERROR SET_X(struct Ripple *, struct Variable *);
static ERROR SET_XOffset(struct Ripple *, struct Variable *);
static ERROR SET_Y(struct Ripple *, struct Variable *);
static ERROR SET_YOffset(struct Ripple *, struct Variable *);

static struct FieldArray RippleFields[] = {
   { "Flags",        0, FDF_LONGFLAGS|FDF_RW, (LONG)&RippleFlags, NULL, NULL },
   { "Strength",     0, FDF_LONG|FDF_RW,      0, NULL, SET_Strength },
   { "Density",      0, FDF_LONG|FDF_RW,      0, NULL, SET_Density },
   { "DropHeight",   0, FDF_LONG|FDF_RW,      0, NULL, SET_DropHeight },
   { "CanvasWidth",  0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "CanvasHeight", 0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   /*** Variable Fields ***/
   { "Height",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

static ERROR RIPPLE_ActionNotify(struct Ripple *, struct acActionNotify *);
static ERROR RIPPLE_Free(struct Ripple *, APTR);
static ERROR RIPPLE_Init(struct Ripple *, APTR);
static ERROR RIPPLE_NewObject(struct Ripple *, APTR);
static ERROR RIPPLE_Timer(struct Ripple *, struct acTimer *);

static struct ActionArray RippleActions[] = {
   { AC_ActionNotify, RIPPLE_ActionNotify },
   { AC_Free,         RIPPLE_Free },
   { AC_Init,         RIPPLE_Init },
   { AC_NewObject,    RIPPLE_NewObject },
   { AC_Timer,        RIPPLE_Timer },
   { NULL, NULL }
};

/****************************************************************************/

#ifndef PI
#define PI 3.14159265358979323846
#endif

#define FSINMAX  2047
#define SINFIX   16
#define FSINBITS 16

#define FSin(x) FSinTab[(x) & FSINMAX]
#define FCos(x) FCosTab[(x) & FSINMAX]

LONG FSinTab[2048];
LONG FCosTab[2048];

static void HeightBlob(struct Ripple *, LONG, LONG y, LONG, LONG, LONG);
static void WarpBlob(struct Ripple *, LONG, LONG, LONG, LONG, LONG);
static void SineBlob(struct Ripple *, LONG, LONG, LONG, LONG, LONG);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   WORD i;
   FLOAT angle;

   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &RenderMod, NULL,
         FID_Name|TSTRING,   "render",
         FID_Version|TFLOAT, MODVERSION_RENDER,
         TAGEND) IS ERR_Okay) {
      if (GetField(RenderMod, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /*** Create fixed sin and cos tables ***/

   for (i=0; i < 2048; i++) {
      angle = (FLOAT)i * (PI/1024.0);
      FSinTab[i] = (LONG)(sin(angle) * (FLOAT)0x10000);
      FCosTab[i] = (LONG)(cos(angle) * (FLOAT)0x10000);
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&RippleClass, NULL,
      FID_BaseClassID|TLONG, ID_RIPPLE,
      FID_Version|TFLOAT,    VER_RIPPLE,
      FID_Name|TSTRING,      "Ripple",
      FID_Category|TLONG,    CCF_EFFECT,
      FID_Actions|TPTR,      RippleActions,
      FID_Fields|TPTR,       RippleFields,
      FID_Size|TLONG,        sizeof(struct Ripple),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (RenderMod)   { Action(AC_Free, RenderMod, NULL);   RenderMod = NULL; }
   if (RippleClass) { Action(AC_Free, RippleClass, NULL); RippleClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Ripple: ActionNotify
*/

void AddBlob(struct Ripple *Self, LONG XCoord, LONG YCoord)
{
   struct Variable value;
   LONG x, y, width, height;

   value.Type = FD_LARGE;
   GET_Width(Self, &value);  width = value.Large;
   GET_Height(Self, &value); height = value.Large;

   if (width < Self->CanvasWidth) x = XCoord;
   else x = XCoord * Self->CanvasWidth / width;

   if (height < Self->CanvasHeight) y = YCoord;
   else y = YCoord * Self->CanvasHeight / height;

   if ((x < 0) OR (y < 0) OR (x > Self->CanvasWidth) OR (y > Self->CanvasHeight)) return;

   if (Self->Buttons & JD_LMB) {
      if (Self->Flags & FLF_PUNCH) SineBlob(Self, x, y, Self->Strength, -Self->DropHeight, Self->Page);
      else if (Self->Flags & FLF_WARP) WarpBlob(Self, x, y, Self->Strength, Self->DropHeight, Self->Page);
      else HeightBlob(Self, x, y, Self->Strength/2, Self->DropHeight, Self->Page);
   }
   else if (Self->Buttons & JD_RMB) {
      if (Self->Flags & FLF_PUNCH)     Self->Flags = (Self->Flags & (~(FLF_WARP|FLF_PUNCH))) | FLF_WARP;
      else if (Self->Flags & FLF_WARP) Self->Flags = Self->Flags & (~(FLF_WARP|FLF_PUNCH));
      else Self->Flags = Self->Flags = (Self->Flags & (~(FLF_WARP|FLF_PUNCH))) | FLF_PUNCH;
   }
}

static ERROR RIPPLE_ActionNotify(struct Ripple *Self, struct acActionNotify *NotifyArgs)
{
   if ((NotifyArgs->ActionID IS AC_UserClick) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct acUserClick *click;
      struct Variable value;
      LONG x, y;

      click = (struct acUserClick *)NotifyArgs->Args;
      Self->Buttons |= click->ButtonFlags;

      value.Type = FD_LARGE;
      GET_X(Self, &value); x = value.Large;
      GET_Y(Self, &value); y = value.Large;

      AddBlob(Self, click->XCoord - x, click->YCoord - y);
   }
   else if ((NotifyArgs->ActionID IS AC_UserMovement) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct acUserMovement *move;
      struct Variable value;
      LONG x, y;

      move = (struct acUserMovement *)NotifyArgs->Args;

      value.Type = FD_LARGE;
      GET_X(Self, &value); x = value.Large;
      GET_Y(Self, &value); y = value.Large;
      AddBlob(Self, move->XCoord - x, move->YCoord - y);
   }
   else if ((NotifyArgs->ActionID IS AC_UserClickRelease) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct acUserClickRelease *release;
      release = (struct acUserClickRelease *)NotifyArgs->Args;

      Self->Buttons &= ~release->ButtonFlags;
   }
   else if ((NotifyArgs->ActionID IS AC_Resize) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct acResize *resize;
      resize = (struct acResize *)NotifyArgs->Args;

      if ((Self->Buffer->Width != resize->Width) OR (Self->Buffer->Height != resize->Height)) {
         Action(AC_Resize, Self->Buffer, resize);
      }

      Self->DrawableWidth = resize->Width;
      Self->DrawableHeight = resize->Height;
   }
   else if ((NotifyArgs->ActionID IS AC_Redimension) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct acResize resize;
      struct acRedimension *redimension;
      redimension = (struct acRedimension *)NotifyArgs->Args;

      resize.Width  = redimension->Width;
      resize.Height = redimension->Height;
      resize.Depth  = 0;
      if ((Self->Buffer->Width != resize.Width) OR (Self->Buffer->Height != resize.Height)) {
         Action(AC_Resize, Self->Buffer, &resize);
      }

      Self->DrawableWidth  = redimension->Width;
      Self->DrawableHeight = redimension->Height;
   }
   else if ((NotifyArgs->ActionID IS AC_Draw) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct Bitmap *bitmap;
      struct mtReadRGBPixel pixel;
      struct RGB rgb;
      struct Variable var;
      OBJECTPTR drawable;
      LONG deltax, deltay, x, y, offset, srcx, srcy, endx, endy;
      LONG yremainder, xcoord, ycoord, width, height;
      WORD xremainder, domain, mapx, startmapx;
      WORD *map, opacity;

      if (AccessObject(Self->DrawableID, 0, &drawable) != ERR_Okay) {
         return(ObjectError(Self, ERH_Draw, ERR_ExclusiveDenied));
      }

      if ((GetField(drawable, FID_Bitmap, FT_POINTER, &bitmap) != ERR_Okay) OR (!bitmap)) {
         ReleaseObject(drawable);
         return(ObjectError(Self, ERH_Draw, ERR_GetField));
      }

      var.Type = FD_LARGE;
      GET_X(Self, &var);      xcoord = var.Large;
      GET_Y(Self, &var);      ycoord = var.Large;
      GET_Width(Self, &var);  width  = var.Large;
      GET_Height(Self, &var); height = var.Large;

      if (width > Self->CanvasWidth) width = Self->CanvasWidth;
      if (height > Self->CanvasHeight) height = Self->CanvasHeight;
      
      acFlush(bitmap);

      if (Self->Flags & FLF_BLEND) {
         /* This technique simply uses alpha blending to apply the ripples.
         ** This method is useful if you want to achieve an affect that looks
         ** more like a vibration or explosion.
         **
         ** This technique also allows us to draw straight to the destination
         ** bitmap rather than a buffer.
         */

         map = Self->HeightMap[Self->Page];
         endx = xcoord + width;
         endy = ycoord + height;
         startmapx = 0;

         if (xcoord < bitmap->ClipLeft) {
            startmapx += (bitmap->ClipLeft - xcoord);
            xcoord = bitmap->ClipLeft;
         }

         if (ycoord < bitmap->ClipTop) {
            map   += (bitmap->ClipTop - ycoord) * Self->CanvasWidth;
            ycoord = bitmap->ClipTop;
         }

         if (endx > bitmap->ClipRight)  endx = bitmap->ClipRight;
         if (endy > bitmap->ClipBottom) endy = bitmap->ClipBottom;

         for (y=ycoord; y < endy; y++) {
            for (x=xcoord, mapx=startmapx; x < endx; x++, mapx++) {
               bitmap->ReadUCRPixel(bitmap, x+bitmap->XOffset, y+bitmap->YOffset, &rgb);

               opacity   = map[mapx]>>1;
               if (opacity < 0) opacity = 0;
               else if (opacity > 255) opacity = 255;
               rgb.Red   = rgb.Red   + (((255 - rgb.Red)   * opacity)>>8);
               rgb.Green = rgb.Green + (((255 - rgb.Green) * opacity)>>8);
               rgb.Blue  = rgb.Blue  + (((255 - rgb.Blue)  * opacity)>>8);

               bitmap->DrawUCRPixel(bitmap, x + bitmap->XOffset, y + bitmap->YOffset, &rgb);
            }
            map += Self->CanvasWidth;
         }
      }
      else if (Self->Flags & FLF_GREYSCALE) {
         /*** This technique draws the height map 'as is' to the destination bitmap ***/

         map = Self->HeightMap[Self->Page];

         if (ycoord < bitmap->ClipTop) {
            height = height - (bitmap->ClipTop - ycoord);
            map   += (bitmap->ClipTop - ycoord) * Self->CanvasWidth;
            ycoord = bitmap->ClipTop;
         }

         if (ycoord + height > bitmap->ClipBottom) height = bitmap->ClipBottom - ycoord;

         for (y=ycoord; y < height; y++) {
            for (x=xcoord, mapx=0; x < width; x++, mapx++) {
               rgb.Red   = map[mapx]>>1;
               if (rgb.Red < 0) rgb.Red = 0; else if (rgb.Red > 255) rgb.Red = 255;
               rgb.Green = rgb.Red;
               rgb.Blue  = rgb.Red;
               bitmap->DrawUCRPixel(bitmap, x+bitmap->XOffset, y+bitmap->YOffset, &rgb);
            }
            map += Self->CanvasWidth;
         }
      }
      else {
         /* This is the standard technique for drawing water ripples.  It uses
         ** an internal buffer for applying the blit.
         */

         map        = Self->HeightMap[Self->Page];
         offset     = Self->CanvasWidth;
         yremainder = 0;
         srcy = 0;
         for (y=0; y < height; y++) {
            xremainder = 0;
            srcx = 0;
            for (x=0; x < width; x++) {
               /*** Get the difference between the current pixel height and the next ***/

               deltax = map[offset] - map[offset+1];
               deltay = map[offset] - map[offset+Self->CanvasWidth];

               /*** Grab the pixel that we are going to use as our source ***/

               pixel.XCoord = srcx + (deltax>>3);
               pixel.YCoord = srcy + (deltay>>3);
               if (pixel.XCoord < 0) pixel.XCoord = 0;
               else if (pixel.XCoord > Self->CanvasWidth-1) pixel.XCoord = Self->CanvasWidth-1;
               if (pixel.YCoord < 0) pixel.YCoord = 0;
               else if (pixel.YCoord > Self->CanvasHeight-1) pixel.YCoord = Self->CanvasHeight-1;

               if ((pixel.XCoord >= bitmap->ClipLeft) AND (pixel.YCoord >= bitmap->ClipTop) AND
                   (pixel.XCoord < bitmap->ClipRight) AND (pixel.YCoord < bitmap->ClipBottom)) {
                  bitmap->ReadUCRPixel(bitmap, pixel.XCoord+bitmap->XOffset, pixel.YCoord+bitmap->YOffset, &rgb);
               }

               /*** Expand to fit the drawable if the canvas is smaller than it ***/

               if (width > Self->CanvasWidth) {
                  xremainder += Self->CanvasWidth;
                  while (xremainder >= width) {
                     offset++;
                     srcx++;
                     xremainder -= width;
                  }
               }
               else { offset++; srcx++; }

               /*** Apply lighting based on the height differential ***/

               rgb.Red   -= deltax;
               rgb.Green -= deltax;
               rgb.Blue  -= deltax;
               if (rgb.Red < 0)   rgb.Red   = 0; else if (rgb.Red   > 255) rgb.Red   = 255;
               if (rgb.Green < 0) rgb.Green = 0; else if (rgb.Green > 255) rgb.Green = 255;
               if (rgb.Blue < 0)  rgb.Blue  = 0; else if (rgb.Blue  > 255) rgb.Blue  = 255;

               /*** Draw the destination pixel to our internal buffer ***/

               Self->Buffer->DrawUCRPixel(Self->Buffer, x, y, &rgb);
            }

            if (width < Self->CanvasWidth) offset += Self->CanvasWidth - width;

            srcy++;

            if (Self->CanvasHeight < height) {
               /*** Expand height ***/
               domain      = height - Self->CanvasHeight;
               yremainder += domain;
               while (yremainder >= Self->CanvasHeight) {
                  yremainder -= Self->CanvasHeight;
                  y++;
                  ActionTags(MT_BlitArea, Self->Buffer, Self->Buffer, NULL, 0, y-1, width, 1, 0, y);
               }
            }
         }

         ActionTags(MT_BlitArea, Self->Buffer, bitmap, NULL, 0, 0, Self->Buffer->Width, Self->Buffer->Height, 0, 0);
      }

      ReleaseObject(drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Ripple: Free
*/

static ERROR RIPPLE_Free(struct Ripple *Self, APTR Void)
{
   UnsubscribeTimer(Self->Head.UniqueID);

   if (Self->HeightMap[0]) {
      FreeMemory(Self->HeightMap[0]);
      Self->HeightMap[0] = NULL;
   }

   if (Self->HeightMap[1]) {
      FreeMemory(Self->HeightMap[1]);
      Self->HeightMap[1] = NULL;
   }

   if (Self->Buffer) {
      Action(AC_Free, Self->Buffer, NULL);
      Self->Buffer = NULL;
   }

   /*** Remove all subscriptions to our Drawable container ***/

   if (Self->DrawableID) {
      OBJECTPTR Drawable;
      if (AccessObject(Self->DrawableID, 3000, &Drawable) IS ERR_Okay) {
         UnsubscribeAction(Drawable, NULL,  GetUniqueID(Self));
         ReleaseObject(Drawable);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Ripple: Init
*/

static ERROR RIPPLE_Init(struct Ripple *Self, APTR Void)
{
   struct Variable value;
   struct RenderInfo info;
   OBJECTPTR Drawable;

   /*** Find the Drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   /* If dimension settings are missing (e.g. if it is impossible to
   ** determine width, height or a coordinate), then w
e set the missing
   ** fields to maximum possible values.
   */

   value.Type = FD_DOUBLE;
   value.Double = 0;

   if (!(Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_X(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_Y(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_XOffset(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_YOffset(Self, &value);
   }

   /*** Allocate the internal Bitmap ***/

   if (GetRenderInfo(Self->DrawableID, &info, sizeof(info)) IS ERR_Okay) {
      Self->DrawableWidth         = info.Width;
      Self->DrawableHeight        = info.Height;
      Self->Buffer->BitsPerPixel  = info.BitsPerPixel;
      Self->Buffer->BytesPerPixel = info.BytesPerPixel;
      Self->Buffer->Width         = Self->DrawableWidth;
      Self->Buffer->Height        = Self->DrawableHeight;
      Self->Buffer->DataFlags    |= MEM_DATA;
      if (Action(AC_Init, Self->Buffer, NULL) != ERR_Okay) {
         return(ObjectError(Self, ERH_Init, ERR_Init));
      }
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   /*** Subscribe to our container's Resize action ***/

   if (AccessObject(Self->DrawableID, 5000, &Drawable) IS ERR_Okay) {
      SubscribeActionTags(Drawable, Self,
         AC_Resize,
         AC_Redimension,
         AC_Draw,
         TAGEND);
      if (Self->Flags & FLF_INTERACTIVE) {
         SubscribeActionTags(Drawable, Self,
            AC_UserClick,
            AC_UserClickRelease,
            AC_UserMovement,
            TAGEND);
      }
      ReleaseObject(Drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   Self->SurferXAngle = RandomNumber(2048);
   Self->SurferYAngle = RandomNumber(2048);
   Self->SwirlAngle   = RandomNumber(2048);

   if (AllocMemory(sizeof(WORD) * (Self->CanvasWidth+1) * (Self->CanvasHeight+4), MEM_DATA, &Self->HeightMap[0], NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_AllocMemory));
   }

   if (AllocMemory(sizeof(WORD) * (Self->CanvasWidth+1) * (Self->CanvasHeight+4), MEM_DATA, &Self->HeightMap[1], NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_AllocMemory));
   }

   /*** Initialise the Timer ***/

   return(SubscribeTimer(Self->Head.UniqueID, 20));
}

/*****************************************************************************
** Ripple: NewObject
*/

static ERROR RIPPLE_NewObject(struct Ripple *Self, APTR Void)
{
   Self->SurferX      = 80;
   Self->SurferY      = 60;
   Self->Density      = 50;
   Self->DropHeight   = 400;
   Self->Strength     = 15;
   Self->CanvasWidth  = 320;
   Self->CanvasHeight = 200;

   return(NewObject(ID_BITMAP, Self->Head.ObjectFlags|NF_CHILD, &Self->Buffer, NULL));
}

/*****************************************************************************
** Action: Timer
*/

static ERROR RIPPLE_Timer(struct Ripple *Self, struct acTimer *Args)
{
   struct Variable value;
   struct acDraw draw;
   LONG newh, x, y, count, density;
   WORD *newmap, *oldmap, *map;

   /*** The surfer ***/

   if (Self->Flags & FLF_SURFER) {
      x = (Self->CanvasWidth/2) + ((((FSin((Self->SurferXAngle * 65)>>8)>>8) * (FSin((Self->SurferXAngle*349)>>8)>>8)) * ((Self->CanvasWidth-8)/2)) >> 16);
      y = (Self->CanvasHeight/2) + ((((FSin((Self->SurferYAngle * 377)>>8)>>8) * (FSin((Self->SurferYAngle*84)>>8)>>8)) * ((Self->CanvasHeight-8)/2)) >> 16);
      Self->SurferXAngle += 13;
      Self->SurferYAngle += 12;

      if (Self->Flags & FLF_PUNCH) {
         SineBlob(Self, (Self->SurferX + x)/2, (Self->SurferY + y)/2, 3, -1200, Self->Page);
         SineBlob(Self, x, y, 4, -2000, Self->Page);
      }
      else HeightBlob(Self, x, y, Self->Strength/2, Self->DropHeight, Self->Page);

      Self->SurferX = x;
      Self->SurferY = y;
   }

   /*** The raindrops ***/

   if (Self->Flags & FLF_RAINDROPS) {
      x = RandomNumber(Self->CanvasWidth-2) + 1;
      y = RandomNumber(Self->CanvasHeight-2) + 1;
      map = Self->HeightMap[Self->Page];
      map[y * Self->CanvasWidth + x] = RandomNumber(Self->DropHeight * 4);
   }

   /*** The big splashes ***/

   if (Self->Flags & FLF_SPLASHES) {
      if (!RandomNumber(20)) {
         x = Self->Strength + RandomNumber(Self->CanvasWidth - (2 * Self->Strength));
         y = Self->Strength + RandomNumber(Self->CanvasHeight - (2 * Self->Strength));

         if (Self->Flags & FLF_PUNCH) SineBlob(Self, x, y, Self->Strength, -Self->DropHeight*6, Self->Page);
         else HeightBlob(Self, x, y, Self->Strength/2, Self->DropHeight, Self->Page);
      }
   }

   /*** Swirling effect ***/

   if (Self->Flags & FLF_SWIRL) {
      x = (Self->CanvasWidth/2) + (((FCos(Self->SwirlAngle)) * (25))>>16);
      y = (Self->CanvasHeight/2) + (((FSin(Self->SwirlAngle)) * (25))>>16);
      Self->SwirlAngle += 50;
      if (Self->Flags & (FLF_PUNCH|FLF_WARP)) SineBlob(Self, x, y, 15, -1200, Self->Page);
      else HeightBlob(Self, x, y, 6, Self->DropHeight, Self->Page);
   }

   /*** Move the water ***/

   newmap = Self->HeightMap[Self->Page^1];
   oldmap = Self->HeightMap[Self->Page];
   density = Self->Density / 10;
   if (Self->Flags & FLF_HIGHQUALITY) {
      count = (2 * Self->CanvasWidth) + 2;
      for (y=2; y < Self->CanvasHeight-2; y++) {
         for (x=2; x < Self->CanvasWidth-2; x++) {
            newh = ((((oldmap[count + Self->CanvasWidth] + oldmap[count - Self->CanvasWidth] + oldmap[count + 1] + oldmap[count - 1])<<1)
                   + ((oldmap[count - Self->CanvasWidth - 1] + oldmap[count - Self->CanvasWidth + 1] + oldmap[count + Self->CanvasWidth - 1] + oldmap[count + Self->CanvasWidth + 1]))
                   + ((oldmap[count - (Self->CanvasWidth*2)] + oldmap[count + (Self->CanvasWidth*2)] + oldmap[count - 2] + oldmap[count + 2])>>1)
                   + ((oldmap[count - (Self->CanvasWidth*2) - 1] + oldmap[count - (Self->CanvasWidth*2) + 1] + oldmap[count + (Self->CanvasWidth*2) - 1] + oldmap[count + (Self->CanvasWidth*2) + 1]
                   + oldmap[count - 2 - Self->CanvasWidth] + oldmap[count - 2 + Self->CanvasWidth] + oldmap[count + 2 - Self->CanvasWidth] + oldmap[count + 2 + Self->CanvasWidth]
                   )>>2))>>3) - (newmap[count]);
            newmap[count] =  newh - (newh >> density);
            count++;
         }
         count += 4;
      }
   }
   else {
      /*** Would be smart to convert this routine to assembler ***/

      count = Self->CanvasWidth + 1;
      for (y = (Self->CanvasHeight+1)*Self->CanvasWidth; count < y; count += 2) {
         for (x = count+Self->CanvasWidth-2; count < x; count++) {
            newh = ((oldmap[count + Self->CanvasWidth] + oldmap[count - Self->CanvasWidth] + oldmap[count + 1] + oldmap[count - 1]
                    + oldmap[count - Self->CanvasWidth - 1] + oldmap[count - Self->CanvasWidth + 1]
                    + oldmap[count + Self->CanvasWidth - 1] + oldmap[count + Self->CanvasWidth + 1])>>2) - newmap[count];
            newmap[count] =  newh - (newh >> density);
         }
      }
   }


   Self->Page ^= 1;

   value.Type = FD_DOUBLE;
   GET_X(Self, &value);      draw.XCoord = value.Double;
   GET_Y(Self, &value);      draw.YCoord = value.Double;
   GET_Width(Self, &value);  draw.Width  = value.Double;
   GET_Height(Self, &value); draw.Height = value.Double;
   DelayMsg(AC_Draw, Self->DrawableID, &draw);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   CanvasHeight
Short:  Defines the height of the ripple effect buffer.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Once it is initialised, the ripple object allocates its own graphical
workspace for generating the effect.  The CanvasWidth and CanvasHeight fields
have default values, but you may want to set the dimensions of the area as it
has a direct effect on the speed and quality of the algorithm.</p>

<p>If the value is high, the quality will increase but the effect will
take longer to generate.  If the value is low then the quality will
decrease but the effect will be generated more quickly.  It is
recommended that the value you specify is between the range of 50 and
400.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   CanvasWidth
Short:  Defines the width of the ripple effect buffer.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Once it is initialised, the ripple object allocates its own graphical
workspace for generating the effect.  The CanvasWidth and CanvasHeight fields
have default values, but you may want to set the dimensions of the area as it
has a direct effect on the speed and quality of the algorithm.</p>

<p>If the value is high, the quality will increase but the effect will
take longer to generate.  If the value is low then the quality will
decrease but the effect will be generated more quickly.  It is
recommended that the value you specify is between the range of 50 and
400.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Density
Short:  The density of the canvas is determined by this field.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>The Density field controls the density of the ripple canvas.  The higher
the density, the quicker the ripples will fade away.  Very low density values
will cause the ripples to die over a long period of time.</p>

<p>The range of this field is between 0 and 100.</p>

###END###

*****************************************************************************/

static ERROR SET_Density(struct Ripple *Self, LONG *Value)
{
   Self->Density = *Value;
   if (Self->Density > 100) Self->Density = 100;
   else if (Self->Density < 0) Self->Density = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   DropHeight
Short:  The height of new droplets is determined by this field.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>The DropHeight has an effect on the height that is applied to new ripples.
This primarily impacts on the lighting and the initial size of new
ripples.  The minimum height is a value of 10 and the maximum is 1000.  You
may find that very low values are useful for more subtle ripple effects.</p>

###END###

*****************************************************************************/

static ERROR SET_DropHeight(struct Ripple *Self, LONG *Value)
{
   if (*Value < 10) Self->DropHeight = 10;
   else if (*Value > 1000) Self->DropHeight = 1000;
   else Self->DropHeight = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Special flags are applied here.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>Special flags supported by the Ripple class can be specified here.  As
there are many options available, try experimenting with the application of
different flags to create all sorts of different effects.</p>

<p>The following flags are responsible for adding automated effects:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">SURFER</td><td>The surfer mode creates a constantly applied ripple that travels around the display using a sine wave.</td></tr>
<tr><td width="10%">RAINDROPS&nbsp;&nbsp;</td><td>This effect creates the impression of raindrops falling on the canvas.</td></tr>
<tr><td width="10%">SPLASHES</td><td>This effect is identical in style to the raindrops effect but uses bigger drops to create the impression of large splashes.</td></tr>
<tr><td width="10%">SWIRL</td><td>This effect creates a swirling pattern that spins around in the centre of the canvas.</td></tr>
</table>

<p>The following flags determine the type of ripples that are applied.  Note
that if neither of these flags are set, the default 'raindrop' type is
used.</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">PUNCH</td><td>This flag provides a ripple effect combined with a distortion.</td></tr>
<tr><td width="10%">WARP</td><td>This flag provides a pure distortion that has no rippling effect at all.</td></tr>
</table>

<p>The following flags impact upon drawing style and user activity:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">BLEND</td><td>This flag changes the drawing routine so that it applies pure alpha blending for the ripples with no other special techniques for 3D styling.</td></tr>
<tr><td width="10%">GREYSCALE</td><td>This flag forces the drawing routine to simply draw the grey scale version of the ripple's internal effect buffer.  Very useful for generating a different type of effect.</td></tr>
<tr><td width="10%">HIGHQUALITY&nbsp;&nbsp;</td><td>Use this flag to enable high-quality mode.  This is slower than the default quality standard.</td></tr>
<tr><td width="10%">INTERACTIVE</td><td>Set this flag to enable user interactive mode.  This allows the user to draw his own ripples using the attached pointing device.  The first button places a ripple and the second button switches the current ripple type that is being used.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Height
Short:  The height of the ripple graphic is specified here.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>If a ripple graphic is to be limited to a specific drawable area, set this
field to limit the clipping height.  A percentage may be specified if the
FD_PERCENT flag is used when setting the field.</p>

###SEE ALSO###
Field: RippleHeight, RippleWidth, Width

###END###

*****************************************************************************/

static ERROR GET_Height(struct Ripple *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   FLOAT containerheight, YCoord, YOffset;

   if (!(Value->Type & (FD_DOUBLE|FD_LARGE))) {
      return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   }

   if (Value->Type & FD_PERCENTAGE) {
      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Height, FT_FLOAT, &containerheight);
            Value->Double = Self->Height / containerheight * 100;
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else {
         if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
                  (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
            if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
               GetField(drawable, FID_Height, FD_FLOAT, &containerheight);
               if (Self->Dimensions & DMF_FIXED_Y) YCoord = Self->YCoord;
               else YCoord = Self->YCoord * 100 / containerheight;
               if (Self->Dimensions & DMF_FIXED_YOFFSET) YOffset = Self->YOffset;
               else YOffset = Self->YOffset * 100 / containerheight;
               Value->Double = (containerheight - YOffset - YCoord) / containerheight * 100;
               ReleaseObject(drawable);
            }
            else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
         }
         else return(ERR_FieldNotSet);
      }
   }
   else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
      Value->Double = Self->Height;
   }
   else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         GetField(drawable, FID_Height, FT_FLOAT, &containerheight);
         Value->Double = Self->Height * containerheight / 100;
         ReleaseObject(drawable);
      }
      else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
   }
   else {
      if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Height, FD_FLOAT, &containerheight);
            GetFields(Self, FID_YCoord|TFLOAT,  &YCoord,
                            FID_YOffset|TFLOAT, &YOffset,
                            TAGEND);
            Value->Double = containerheight - YOffset - YCoord;
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else return(ERR_FieldNotSet);
   }

   Value->Large = Value->Double;
   return(ERR_Okay);
}

static ERROR SET_Height(struct Ripple *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Ripple] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
      Self->Height     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Ripple] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
      Self->Height     = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Strength
Short:  The amount of strength applied to new ripples.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>The Strength field determines the amount of strength that is applied to
new ripples.  This primarily impacts on the overall size of new ripples, so
if your ripples are looking too small, it will help to increase the Strength
field.  Values between 10 and 25 are usually appropriate.</p>

<p>The minimum setting for this field is 2 and the maximum is 200.</p>

###END###

*****************************************************************************/

static ERROR SET_Strength(struct Ripple *Self, LONG *Value)
{
   if (*Value < 2) Self->Strength = 2;
   else if (*Value > 200) Self->Strength = 200;
   else Self->Strength = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  The width of the ripple graphic is specified here.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>If a ripple graphic is to be limited to a specific drawable area, set this
field to limit the clipping width.  A percentage may be specified if the
FD_PERCENT flag is used when setting the field.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

static ERROR GET_Width(struct Ripple *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   FLOAT ContainerWidth, XCoord, XOffset;

   if (!(Value->Type & (FD_DOUBLE|FD_LARGE))) {
      return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   }

   if (Value->Type & FD_PERCENTAGE) {
      if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_FIXED_WIDTH) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width / ContainerWidth * 100;
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else {
         if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
                  (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
            if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
               GetField(drawable, FID_Width, FD_FLOAT, &ContainerWidth);
               if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
               else XCoord = Self->XCoord * 100 / ContainerWidth;
               if (Self->Dimensions & DMF_FIXED_XOFFSET) XOffset = Self->XOffset;
               else XOffset = Self->XOffset * 100 / ContainerWidth;
               Value->Double = (ContainerWidth - XOffset - XCoord) * 100 / ContainerWidth;
               ReleaseObject(drawable);
            }
            else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
         }
         else return(ERR_FieldNotSet);
      }
   }
   else if (Self->Dimensions & DMF_FIXED_WIDTH) {
      Value->Double = Self->Width;
   }
   else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         GetField(drawable, FID_Width, FT_FLOAT, &ContainerWidth);
         Value->Double = Self->Width * ContainerWidth / 100;
         ReleaseObject(drawable);
      }
      else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
   }
   else {
      if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = ContainerWidth * Self->XCoord / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - XCoord - Self->XOffset;
            else Value->Double = ContainerWidth - XCoord - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else return(ERR_Failed);
   }

   Value->Large = Value->Double;
   return(ERR_Okay);
}

static ERROR SET_Width(struct Ripple *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Ripple] A width setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
      Self->Width      = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Ripple] A width setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
      Self->Width      = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  Defines the horizontal position of the ripple graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of an image can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Fields: Align, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Ripple *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_X) value = Self->XCoord;
      else if (Self->Dimensions & DMF_RELATIVE_X) {
         value = Self->XCoord * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - width - Self->XOffset;
         else value = containerwidth - width - (containerwidth * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Ripple *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Defines the horizontal offset of the ripple graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the image will be drawn
from that X coordinate up to the width of the container, minus the value
given in the XOffset.  This means that the width of the image is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the image will be drawn at an X coordinate calculated from the formula
"XCoord = ContainerWidth - RippleWidth - XOffset".</p>

###SEE ALSO###
Fields: Align, XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Ripple *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_XOFFSET) value = Self->XOffset;
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         value = Self->XOffset * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_X) value = containerwidth - (Self->XCoord + width);
         else value = containerwidth - ((Self->XCoord * containerwidth / 100) + width);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Ripple *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XOffset = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XOffset = Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  Defines the vertical position of the ripple graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of an image can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Fields: Align, XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_Y(struct Ripple *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_Y) value = Self->YCoord;
      else if (Self->Dimensions & DMF_RELATIVE_Y) {
         value = Self->YCoord * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - height - Self->YOffset;
         else value = containerheight - height - (containerheight * Self->YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Ripple *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Defines the vertical offset of the ripple graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the image will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the ripple is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the image will be drawn at a Y coordinate calculated from the formula
"YCoord = ContainerHeight - Height - YOffset".</p>

###SEE ALSO###
Fields: Align, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Ripple *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   value = 0;
   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_YOFFSET) value = Self->YOffset;
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         value = Self->YOffset * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_Y) value = containerheight - (Self->YCoord + height);
         else value = containerheight - ((Self->YCoord * containerheight / 100) + height);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Ripple *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->YOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************
** Function: HeightBlob()
*/

void HeightBlob(struct Ripple *Self, LONG x, LONG y, LONG radius, LONG height, LONG page)
{
   LONG rquad, cx, cy, cyq, left, top, right, bottom;
   WORD *map;

   left   = -radius;
   right  = radius;
   top    = -radius;
   bottom = radius;

   /*** Perform edge clipping ***/

   if (x - radius < 1) left -= (x-radius-1);
   if (y - radius < 1) top  -= (y-radius-1);
   if (x + radius > Self->CanvasWidth-1) right -= (x+radius-Self->CanvasWidth+1);
   if (y + radius > Self->CanvasHeight-1) bottom-= (y+radius-Self->CanvasHeight+1);

   rquad = radius * radius;
   map = Self->HeightMap[page];
   for (cy = top; cy < bottom; cy++) {
      cyq = cy * cy;
      for (cx = left; cx < right; cx++) {
         if (cx * cx + cyq < rquad) {
            map[Self->CanvasWidth * (cy+y) + (cx+x)] += height;
         }
      }
   }
}

/*****************************************************************************
** Function: WarpBlob()
*/

void WarpBlob(struct Ripple *Self, LONG x, LONG y, LONG radius, LONG height, LONG page)
{
   LONG cx, cy, left, top, right, bottom, square, radsquare;
   WORD *map;

   height /= 64;
   left   = -radius;
   right  = radius;
   top    = -radius;
   bottom = radius;

   /*** Perform edge clipping ***/

   if (x - radius < 1) left -= (x-radius-1);
   if (y - radius < 1) top  -= (y-radius-1);
   if (x + radius > Self->CanvasWidth-1) right -= (x+radius-Self->CanvasWidth+1);
   if (y + radius > Self->CanvasHeight-1) bottom-= (y+radius-Self->CanvasHeight+1);

   radsquare = radius * radius;
   map = Self->HeightMap[page];
   for (cy = top; cy < bottom; cy++) {
      for (cx = left; cx < right; cx++) {
         square = cy*cy + cx*cx;
         if (square < radsquare) {
            map[Self->CanvasWidth*(cy+y) + cx+x] += (radius-sqrt(square)) * (FLOAT)(height);
         }
      }
   }
}

/*****************************************************************************
** Function: SineBlob()
*/

void SineBlob(struct Ripple *Self, LONG x, LONG y, LONG radius, LONG height, LONG page)
{
   LONG cx, cy, left, top, right, bottom, square, dist, radsquare;
   WORD *map;
   FLOAT length;

   left   = -radius;
   right  = radius;
   top    = -radius;
   bottom = radius;

   /*** Perform edge clipping ***/

   if (x - radius < 1) left -= (x-radius-1);
   if (y - radius < 1) top  -= (y-radius-1);
   if (x + radius > Self->CanvasWidth-1) right -= (x+radius-Self->CanvasWidth+1);
   if (y + radius > Self->CanvasHeight-1) bottom-= (y+radius-Self->CanvasHeight+1);

   length = (1024.0/(FLOAT)radius) * (1024.0/(FLOAT)radius);
   radsquare = radius * radius;
   map = Self->HeightMap[page];
   for (cy = top; cy < bottom; cy++) {
      for (cx = left; cx < right; cx++) {
         square = cy*cy + cx*cx;
         if (square < radsquare) {
            dist = sqrt(square*length);
            map[Self->CanvasWidth*(cy+y) + cx+x] += (LONG)((FCos(dist)+0xffff)*height)>>19;
         }
      }
   }
}
