/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Keyboard
Version:   2.1
Status:    Stable
ID:        ID_KEYBOARD
Category:  I/O
Date:      August 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 1996-2004.  All rights reserved.
Short:     The keyboard class is used to manage keyboard I/O.
Keywords:  keyboard, key, input, output

###DESCRIPTION###
<p>The Keyboard class is a dummy class type that is used to create objects
that represent key-based hardware devices.  This means that Keyboard objects
don't actually read hardware, they just receive input from the devices
responsible for controlling keyboard I/O.  Objects that subscribe to a
Keyboard object's data channel can then receive this input in a format that
is standardised by the Pandora Engine.</p>

<p>The most common keyboard object is the 'SystemKeyboard', which is
created during the boot process.  This particular object represents the
main PC keyboard.  If you are designing a system in which you want multiple
keyboards, you can create as many keyboard objects as you have hardware or
virtual devices to associate with them.  This can be useful for hooking up
remote keyboards over a network, or attaching extra devices that have their
own keyboard to a PC.</p>

<p>The Keyboard class supports translation methods for converting raw
key values into characters and vice versa.  Please use them rather than
translating the raw key values yourself, or you will lose vital keymap
translations that may be required on behalf of the user.</p>

###END###

*****************************************************************************/

#ifdef __linux__
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/kd.h>
#endif

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/input/all.h>
#include <pandora/data/config.h>
#include <pandora/files/file.h>
#include <pandora/functions/strtoint.c>
#include <pandora/functions/strcopy.c>
#include <pandora/functions/writeutf8.c>
#include <pandora/input/keymaptable.h>
#include <pandora/modules/strings.h>

#define UNICODE_TABLE
#include <pandora/input/unicodetable.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_KEYBOARD 2.1

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_KEYBOARD, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1996-2004.  All rights reserved.",
   "August 2004",
   "Keyboard"
};

static struct KernelBase *KernelBase = NULL;
static struct StringsBase *StringsBase = NULL;
static OBJECTPTR KeyboardClass = NULL;
static OBJECTPTR modStrings = NULL;
static OBJECTID glBroadcastID = 0;

#ifdef __linux__
static UBYTE glKeyConversion[128] = {
 /* 0x00 */  NULL, K_ESCAPE, K_1, K_2, K_3, K_4, K_5, K_6, K_7, K_8, K_9, K_0, K_MINUS, K_EQUALS, K_BACKSPACE, K_TAB,
 /* 0x10 */  K_Q, K_W, K_E, K_R, K_T, K_Y, K_U, K_I, K_O, K_P, K_LSQUARE, K_RSQUARE, K_ENTER, K_LCONTROL, K_A, K_S,
 /* 0x20 */  K_D, K_F, K_G, K_H, K_J, K_K, K_L, K_SEMICOLON, K_APOSTROPHE, K_REVERSEQUOTE, K_LSHIFT, K_BACKSLASH, K_Z, K_X, K_C, K_V,
 /* 0x30 */  K_B, K_N, K_M, K_COMMA, K_DOT, K_SLASH, K_RSHIFT, K_NPMULTIPLY, K_LALT, K_SPACE, K_CAPSLOCK, K_F1, K_F2, K_F3, K_F4, K_F5,
 /* 0x40 */  K_F6, K_F7, K_F8, K_F9, K_F10, K_NUMLOCK, K_SCROLL, K_NP7, K_NP8, K_NP9, K_NPMINUS, K_NP4, K_NP5, K_NP6, K_NPPLUS, K_NP1,
 /* 0x50 */  K_NP2, K_NP3, K_NP0, K_NPDOT, K_SYSRQ, NULL, K_LESSGREATER, K_F11, K_F12, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
 /* 0x60 */  K_NPENTER, K_RCONTROL, K_NPDIVIDE, K_PRTSCR, K_RALT, K_BREAK, K_HOME, K_UP, K_PAGEUP, K_LEFT, K_RIGHT, K_END, K_DOWN, K_PAGEDOWN, K_INSERT, K_DELETE,
 /* 0x70 */  K_MACRO, K_F13, K_F14, K_HELP, K_REDO, K_F17, NULL, K_PAUSE, NULL, NULL, NULL, NULL, NULL, K_LCOMMAND, K_RCOMMAND, K_MENU
};

static UBYTE glRepeatKey[256] = { 0 };
static LONG glKeyFlags = NULL;
static LONG glLEDFlags = NULL;
#endif

struct deadchar {
   UWORD unicode;
   UWORD conversion;
};

struct deadmap {
   UWORD deadkey;
   struct deadchar *convert;
};

struct deadchar conv_grave[] = {
   { 'A', 0x00c0 }, { 'E', 0x00c8 }, { 'I', 0x00cc }, { 'O', 0x00d2 }, { 'U', 0x00d9 },
   { 'a', 0x00e0 }, { 'e', 0x00e8 }, { 'i', 0x00ec }, { 'o', 0x00f2 }, { 'u', 0x00f9 },
   { 0, 0 }
};

struct deadchar conv_acute[] = {
   { 'A', 0x00c1 }, { 'E', 0x00c9 }, { 'I', 0x00cd }, { 'O', 0x00d3 }, { 'U', 0x00da }, { 'C', 0x0106 }, { 'L', 0x0139 }, { 'N', 0x0143 }, { 'R', 0x0154 }, { 'S', 0x015a }, { 'Z', 0x0179 },
   { 'a', 0x00e1 }, { 'e', 0x00e9 }, { 'i', 0x00ed }, { 'o', 0x00f3 }, { 'u', 0x00fa }, { 'c', 0x0107 }, { 'l', 0x013a }, { 'n', 0x0144 }, { 'r', 0x0155 }, { 's', 0x015b }, { 'z', 0x017a },
   { 0, 0 }
};

struct deadchar conv_circumflex[] = {
   { 'A', 0x00c2 }, { 'E', 0x00ca }, { 'I', 0x00ce }, { 'O', 0x00d4 }, { 'U', 0x00db }, { 'C', 0x0108 }, { 'G', 0x011c }, { 'H', 0x0124 }, { 'J', 0x0134 }, { 'S', 0x015c }, { 'W', 0x0174 }, { 'Y', 0x0176 },
   { 'a', 0x00e2 }, { 'e', 0x00ea }, { 'i', 0x00ee }, { 'o', 0x00f4 }, { 'u', 0x00fb }, { 'c', 0x0109 }, { 'g', 0x011d }, { 'h', 0x0125 }, { 'j', 0x0135 }, { 's', 0x015d }, { 'w', 0x0175 }, { 'y', 0x0177 },
   { 0, 0 }
};

struct deadchar conv_tilde[] = {
   { 'A', 0x00c2 }, { 'N', 0x00d1 }, { 'O', 0x00d5 }, { 'I', 0x0128 }, { 'U', 0x0168 },
   { 'a', 0x00e3 }, { 'n', 0x00f1 }, { 'o', 0x00f5 }, { 'i', 0x0129 }, { 'u', 0x0169 },
   { 0, 0 }
};

struct deadchar conv_dieresis[] = {
   { 'A', 0x00c4 }, { 'E', 0x00cb }, { 'I', 0x00cf }, { 'O', 0x00d6 }, { 'U', 0x00dc },
   { 'a', 0x00e4 }, { 'e', 0x00eb }, { 'i', 0x00ef }, { 'o', 0x00f6 }, { 'u', 0x00fc },
   { 0, 0 }
};

struct deadchar conv_macron[] = {
   { 'A', 0x0100 }, { 'E', 0x0112 }, { 'I', 0x012a }, { 'O', 0x014c }, { 'U', 0x016a },
   { 'a', 0x0101 }, { 'e', 0x0113 }, { 'i', 0x012b }, { 'o', 0x014d }, { 'u', 0x016b },
   { 0, 0 }
};

struct deadchar conv_breve[] = {
   { 'A', 0x0102 }, { 'E', 0x0114 }, { 'G', 0x011e }, { 'I', 0x012c }, { 'O', 0x014e }, { 'U', 0x016c },
   { 'a', 0x0103 }, { 'e', 0x0115 }, { 'g', 0x011f }, { 'i', 0x012d }, { 'o', 0x014f }, { 'u', 0x016d },
   { 0, 0 }
};

struct deadchar conv_dot[] = {
   { 'C', 0x010a }, { 'E', 0x0116 }, { 'G', 0x0120 }, { 'I', 0x0130 }, { 'Z', 0x017b }, { 'L', 0x013f },
   { 'c', 0x010b }, { 'e', 0x0117 }, { 'g', 0x0121 }, { 'i', 0x0131 }, { 'z', 0x017c }, { 'l', 0x0140 },
   { 0, 0 }
};

struct deadchar conv_ring[] = {
   { 'A', 0x00c5 }, { 'U', 0x016e },
   { 'a', 0x00e5 }, { 'u', 0x016f },
   { 0, 0 }
};

struct deadchar conv_caron[] = {
   { 'C', 0x010c }, { 'D', 0x010e }, { 'E', 0x011a }, { 'L', 0x013d }, { 'N', 0x0147 }, { 'R', 0x0158 }, { 'S', 0x0160 }, { 'T', 0x0164 }, { 'Z', 0x017d },
   { 'c', 0x010d }, { 'd', 0x010f }, { 'e', 0x011b }, { 'l', 0x013e }, { 'n', 0x0148 }, { 'r', 0x0159 }, { 's', 0x0161 }, { 't', 0x0165 }, { 'z', 0x017e },
   { 0, 0 }
};

struct deadchar conv_cedilla[] = {
   { 'C', 0x00c7 }, { 'G', 0x0122 }, { 'K', 0x0136 }, { 'L', 0x013b }, { 'N', 0x0145 }, { 'R', 0x0156 }, { 'S', 0x015e }, { 'T', 0x0162 },
   { 'c', 0x00e7 }, { 'g', 0x0123 }, { 'k', 0x0137 }, { 'l', 0x013c }, { 'n', 0x0146 }, { 'r', 0x0157 }, { 's', 0x015f }, { 't', 0x0163 },
   { 0, 0 }
};

struct deadchar conv_ogonek[] = {
   { 'A', 0x0104 }, { 'E', 0x0118 }, { 'I', 0x012e }, { 'U', 0x0172 },
   { 'a', 0x0105 }, { 'e', 0x0119 }, { 'i', 0x012f }, { 'u', 0x0173 },
   { 0, 0 }
};

static struct deadmap deadkeys[] = {
   /* 0x02b0 - 0x02ff are non-spacing Unicode characters.
   ** 0x0300 - 0x036f are combining diacritical marks
   */
   { 0x300, conv_grave },      /* grave */
   { 0x301, conv_acute },      /* acute */
   { 0x302, conv_circumflex }, /* circumflex */
   { 0x303, conv_tilde },      /* tilde */
   { 0x304, conv_macron },     /* macron */
   { 0x306, conv_breve },      /* breve */
   { 0x307, conv_dot },        /* dotaccent */
   { 0x308, conv_dieresis },   /* dieresis */
   { 0x30a, conv_ring },       /* ring */
   { 0x30c, conv_caron },      /* caron hacek */
   { 0x327, conv_cedilla },    /* cedilla */
   { 0x328, conv_ogonek },     /* ogonek */
};

/*****************************************************************************
** Class definition
*/

#define LEN_LAYOUT 30

struct Keyboard {
   OBJECT_HEADER
   LONG XShift;
   LONG DeadKeys;

   PRIVATE_FIELDS
   UBYTE prvLayout[LEN_LAYOUT];    /* Name of the layout */
   UWORD prvKeymap[K_LIST_END][4]; /* Normal, +Shift, +Command, +(Shift & Command) */
   FLOAT RepeatRate;
   LONG  RepeatDelay;
   UWORD prvDeadKey;               /* The unicode value of the most recently pressed dead key (permanently held) */
   UWORD prvDeadActive;            /* If a dead key is active, it will be set here (not permanent, usually zero) */
};

static ERROR GET_Layout(struct Keyboard *, STRING *);
static ERROR GET_RepeatDelay(struct Keyboard *, LONG *);
static ERROR GET_RepeatRate(struct Keyboard *, FLOAT *);
static ERROR GET_UnicodeTable(struct Keyboard *, struct UnicodeDescriptor **);

static ERROR SET_Layout(struct Keyboard *, STRING);
static ERROR SET_RepeatDelay(struct Keyboard *, LONG *);
static ERROR SET_RepeatRate(struct Keyboard *, FLOAT *);
static ERROR SET_XShift(struct Keyboard *, LONG *);

enum {
   XSF_COMMAND=0,
   XSF_ALTGR
};

static struct FieldDef XShiftLookup[] = {
   { "Command", XSF_COMMAND },
   { "AltGr",   XSF_ALTGR },
   { NULL, NULL }
};

static struct FieldArray KeyboardFields[] = {
   { "XShift",       0, FDF_LONG|FDF_LOOKUP|FDF_RW,   (LONG)&XShiftLookup, NULL, SET_XShift },
   { "DeadKeys",     0, FDF_LONG|FDF_RW,   0, NULL, NULL },
   /*** Virtual fields ***/
   { "Layout",       0, FDF_STRING|FDF_RW, 0, GET_Layout, SET_Layout },
   { "UnicodeTable", 0, FDF_POINTER|FDF_R, 0, GET_UnicodeTable, NULL },
   { "RepeatRate",   0, FDF_FLOAT|FDF_RW,  0, GET_RepeatRate, SET_RepeatRate },
   { "RepeatDelay",  0, FDF_LONG|FDF_RW,   0, GET_RepeatDelay, SET_RepeatDelay },
   END_FIELD
};

/*****************************************************************************
** Keyboard actions.
*/

static ERROR KEYBOARD_Activate(struct Keyboard *, APTR);
static ERROR KEYBOARD_ClosingTag(struct Keyboard *, APTR);
static ERROR KEYBOARD_DataChannel(struct Keyboard *, struct acDataChannel *);
static ERROR KEYBOARD_EventMessage(struct Keyboard *, struct acEventMessage *);
static ERROR KEYBOARD_Free(struct Keyboard *, APTR);
static ERROR KEYBOARD_Init(struct Keyboard *, APTR);
static ERROR KEYBOARD_NewObject(struct Keyboard *, APTR);
static ERROR KEYBOARD_Reset(struct Keyboard *, APTR);
static ERROR KEYBOARD_SaveToObject(struct Keyboard *, struct acSaveToObject *);
#ifdef __linux__
static ERROR KEYBOARD_Timer(struct Keyboard *, struct acTimer *);
#endif

static struct ActionArray KeyboardActions[] = {
   { AC_Activate,     KEYBOARD_Activate },
   { AC_ClosingTag,   KEYBOARD_ClosingTag },
   { AC_DataChannel,  KEYBOARD_DataChannel },
   { AC_EventMessage, KEYBOARD_EventMessage },
   { AC_Free,         KEYBOARD_Free },
   { AC_Init,         KEYBOARD_Init },
   { AC_NewObject,    KEYBOARD_NewObject },
   { AC_Reset,        KEYBOARD_Reset },
   { AC_SaveToObject, KEYBOARD_SaveToObject },
#ifdef __linux__
   { AC_Timer,        KEYBOARD_Timer },
#endif
   { NULL, NULL }
};

/*****************************************************************************
** Keyboard methods.
*/

static ERROR KEYBOARD_TranslateChars(struct Keyboard *, struct mtTranslateChars *);
static ERROR KEYBOARD_TranslateKeys(struct Keyboard *, struct mtTranslateKeys *);

static struct FunctionField argsTranslateKeys[]  = { { "Keys", ARG_PTR }, { "String", ARG_STR }, { "Size", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsTranslateChars[] = { { "String", ARG_STR }, { "Keys", ARG_PTR }, { "Size", ARG_LONG }, { NULL, NULL } };

static struct MethodArray KeyboardMethods[] = {
   { MT_TranslateKeys,  KEYBOARD_TranslateKeys,  "TranslateKeys",  argsTranslateKeys,  sizeof(struct mtTranslateKeys) },
   { MT_TranslateChars, KEYBOARD_TranslateChars, "TranslateChars", argsTranslateChars, sizeof(struct mtTranslateChars) },
   { NULL, NULL, NULL, NULL }
};

/****************************************************************************/

static ERROR LoadKeymap(struct Keyboard *);
static LONG isdeadkey(UWORD);
static UWORD convertdeadkey(UWORD, UWORD);

/****************************************************************************/

static void ReadLong(OBJECTPTR File, LONG *Value)
{
   struct mtReadFileLong args;
   Action(MT_ReadFileLong, File, &args);
   if (Value) *Value = args.Result;
}

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
         FID_Name|TSTRING,   "strings",
         FID_Version|TFLOAT, MODVERSION_STRINGS,
         TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   FastFindObject("SystemBroadcaster", ID_BROADCAST, &glBroadcastID, 1, NULL);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&KeyboardClass, NULL,
      FID_BaseClassID|TLONG, ID_KEYBOARD,
      FID_Version|TFLOAT,    VER_KEYBOARD,
      FID_Name|TSTRING,      "Keyboard",
      FID_Category|TLONG,    CCF_IO,
      FID_Fields|TPTR,       KeyboardFields,
      FID_Actions|TPTR,      KeyboardActions,
      FID_Methods|TPTR,      KeyboardMethods,
      FID_Size|TLONG,        sizeof(struct Keyboard),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (modStrings)    { acFree(modStrings);    modStrings = NULL; }
   if (KeyboardClass) { acFree(KeyboardClass); KeyboardClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Action: Activate
*/

static ERROR KEYBOARD_Activate(struct Keyboard *Self, APTR Void)
{
#ifdef __linux__
   if (GetResource(RES_KEYBOARDFD) != -1) {

      /*** Set LED status ***/

      SetResource(RES_PRIVILEGEDUSER, TRUE);
      ioctl(GetResource(RES_CONSOLEFD), KDGETLED, &glLEDFlags);
      glLEDFlags = glLEDFlags & 0x0f;
      SetResource(RES_PRIVILEGEDUSER, FALSE);

      SubscribeTimer(Self->Head.UniqueID, 20);
   }

#endif
   return(ERR_Okay);
}

/*****************************************************************************
** Action: ClosingTag
*/

static ERROR KEYBOARD_ClosingTag(struct Keyboard *Self, APTR Void)
{
   acActivate(Self);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  DataChannel
Short: This action can be used to send fake keyboard input.

###DESCRIPTION###
<p>Keypresses can be sent to a Keyboard object through the DataChannel action
by using the DATA_KEYPRESS data type.  All characters that are sent to a
Keyboard object through the data channel will inevitably be reported to
programs that read the Keyboard through the data channel service.</p>

<p>Please note that it is extremely important that the key qualifiers are set
correctly and that each key press is matched with a key release.  If the data
is not defined correctly, the objects receiving the data may not interpret it
as intended.</p>

###RESULT###
ERR_Okay:         The input was successfully processed.
ERR_Args:         Invalid arguments were specified.
ERR_AccessMemory: Access to the internal channel list was denied.

###END###

*****************************************************************************/

static ERROR KEYBOARD_DataChannel(struct Keyboard *Self, struct acDataChannel *Args)
{
   struct dcKeyEntry *src;
   struct acDataChannel channel;
   struct ChannelSubscription *list;
   MEMORYID ChannelListMID;
   ERROR error;
   LONG i, flags, key;
   UWORD keychar;

   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if (GetField(Self, FID_ChannelList, FT_LONG, &ChannelListMID) != ERR_Okay) return(ERR_Okay);

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct dcKeyEntry array[Args->TotalEntries];

      /* Send all the data directly to all objects that have opened data
      ** channels with us.
      */

      channel.ObjectID     = GetUniqueID(Self);
      channel.DataType     = DATA_KEYPRESS;
      channel.Version      = 1;
      channel.Buffer       = array;
      channel.Size         = Args->TotalEntries * sizeof(struct dcKeyEntry);
      channel.TotalEntries = Args->TotalEntries;

      src = Args->Buffer;
      for (i=0; i < Args->TotalEntries; i++) {

         /* Determine whether or not the keypress will translate to a printable
         ** character if TranslateKeys() is called on it.
         */

         flags = src[i].Flags | KQ_NOTPRINTABLE;
         key   = src[i].Value;
         keychar = 0;
         if (key < K_LIST_END) {
            if (flags & KQ_CONTROL) goto next;

            if (Self->XShift IS XSF_ALTGR) {
               if (flags & (KQ_LALT|KQ_COMMAND)) goto next;
            }
            else if (flags & KQ_ALT) goto next;

            if ((Self->XShift IS XSF_ALTGR) AND (flags & KQ_RALT)) {
               if (flags & KQ_SHIFT) keychar = Self->prvKeymap[key][3];
               else keychar = Self->prvKeymap[key][2];
            }
            else if ((Self->XShift IS XSF_COMMAND) AND (flags & KQ_COMMAND)) {
               if (flags & KQ_SHIFT) keychar = Self->prvKeymap[key][3];
               else keychar = Self->prvKeymap[key][2];
            }
            else if (flags & KQ_SHIFT) keychar = Self->prvKeymap[key][1];
            else keychar = Self->prvKeymap[key][0];

            if (keychar) {
               if ((Self->DeadKeys) AND (isdeadkey(keychar))) {
                  /* If the key is a dead-key, it won't be printable until the
                  ** next key is pressed.  We also record the status of the dead-key
                  ** here.
                  */
                  Self->prvDeadKey    = keychar;
                  Self->prvDeadActive = keychar;
               }
               else flags &= ~KQ_NOTPRINTABLE;
            }
         }

         if ((Self->prvDeadActive) AND (!(flags & (KQ_NOTPRINTABLE|KQ_RELEASED)))) {
            flags |= KQ_DEADKEY;
            Self->prvDeadActive = 0;
         }
next:
         array[i].Flags = flags;
         array[i].Value = key;
      }

      if (ChannelListMID) {
         if (AccessMemory(ChannelListMID, MEM_READ, (APTR *)&list) IS ERR_Okay) {
            i = 0;
            while (list[i].SubscriberID) {
               error = ActionMsgPort(AC_DataChannel, list[i].SubscriberID, &channel, list[i].MessagePortMID, list[i].ClassID);
               if (error IS ERR_NoMatchingObject) {
                  DPrintF("DataChannel:","[Keyboard:%d] Subscriber #%d no longer exists, removing from channel list.", Self->Head.UniqueID, list[i].SubscriberID);
                  if (UnsubscribeChannel(Self, list[i].SubscriberID) != ERR_Okay) i++;
               }
               else if (error != ERR_Okay) {
                  DPrintF("@DataChannel:","[Keyboard] Error %d on subscriber %d.", error, list[i].SubscriberID);
                  i++;
               }
               else i++;
            }
            ReleaseMemory(list);
         }
         else {
            DPrintF("@DataChannel:","[Keyboard] Failed to gain access to memblock #%d.", ChannelListMID);
            return(ERR_AccessMemory);
         }
      }
   }
   else if (Args->DataType IS DATA_TEXT) {
      DPrintF("@DataChanel:","[Keyboard] Text data is not yet supported.");
   }
   else DPrintF("@DataChannel:","[Keyboard] Unrecognised data type %d, version %d.", Args->DataType, Args->Version);

   return(ERR_Okay);
}

/*****************************************************************************
** Action: EventMessage
*/

static ERROR KEYBOARD_EventMessage(struct Keyboard *Self, struct acEventMessage *Args)
{
   OBJECTPTR config;
   STRING str;
   DOUBLE fvalue;
   LONG value;
   WORD i;

   if (Args->EventID IS EVT_UserLogin) {
      /*** Reset the keyboard for the new user login ***/

      for (i=0; "American - US101"[i]; i++) Self->prvLayout[i] = "American - US101"[i];
      Self->prvLayout[i] = 0;

      if (CreateObject(ID_CONFIG, NF_CHILD, (OBJECTPTR *)&config, NULL,
         FID_Location|TSTRING, "user:config/keyboard",
         TAGEND) IS ERR_Okay) {

         if (cfgRead(config, "KEYBOARD", "Layout", &str) IS ERR_Okay)             SetFields(Self, FID_Layout|TSTRING, str, TAGEND);
         if (cfgReadFloat(config, "KEYBOARD", "RepeatRate", &fvalue) IS ERR_Okay) SetFields(Self, FID_RepeatRate|TFLOAT, fvalue, TAGEND);
         if (cfgReadInt(config, "KEYBOARD", "RepeatDelay", &value) IS ERR_Okay)   SetFields(Self, FID_RepeatDelay|TLONG, value, TAGEND);
         if (cfgReadInt(config, "KEYBOARD", "XShift", &value) IS ERR_Okay)        SetFields(Self, FID_XShift|TLONG, value, TAGEND);

         acFree(config);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Free
*/

static ERROR KEYBOARD_Free(struct Keyboard *Self, APTR Void)
{
   OBJECTPTR broadcast;

   if (AccessObject(glBroadcastID, 5000, &broadcast) IS ERR_Okay) {
      mtUnsubscribeEvent(broadcast, Self->Head.UniqueID, EVT_UserLogin);
      ReleaseObject(broadcast);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Init
*/

static ERROR KEYBOARD_Init(struct Keyboard *Self, APTR Void)
{
   OBJECTPTR broadcast;

   LoadKeymap(Self);

   if (AccessObject(glBroadcastID, 5000, &broadcast) IS ERR_Okay) {
      mtSubscribeEvent(broadcast, Self->Head.UniqueID, EVT_UserLogin);
      ReleaseObject(broadcast);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: NewObject
*/

static ERROR KEYBOARD_NewObject(struct Keyboard *Self, APTR Void)
{
   struct mtReadConfigInt readint;
   struct mtReadConfigFloat readflt;
   struct mtReadConfig read;
   OBJECTPTR config;
   WORD i;

   Self->RepeatRate  = 10.9;
   Self->RepeatDelay = 250;
   Self->DeadKeys    = TRUE; /* Enable deadkeys by default */

   /*** Set the default layout to US101 ***/

   for (i=0; "American - US101"[i]; i++) Self->prvLayout[i] = "American - US101"[i];
   Self->prvLayout[i] = 0;

   /* Attempt to get the user's preferred layout
   ** from the user:config/keyboard file.
   */

   if (CreateObject(ID_CONFIG, NF_CHILD, (OBJECTPTR *)&config, NULL,
      FID_Location|TSTRING, "user:config/keyboard",
      TAGEND) IS ERR_Okay) {

      read.Section = "KEYBOARD";
      read.Item = "Layout";
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
         SetField(Self, FID_Layout, FT_STRING, read.Data);
      }

      readflt.Section = "KEYBOARD";
      readflt.Item = "RepeatRate";
      if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) Self->RepeatRate = readflt.Float;

      readint.Section = "KEYBOARD";
      readint.Item = "RepeatDelay";
      if (Action(MT_ReadConfigInt, config, &readint) IS ERR_Okay) Self->RepeatDelay = readint.Integer;

      readint.Item = "XShift";
      if (Action(MT_ReadConfigInt, config, &readint) IS ERR_Okay) Self->XShift = readint.Integer;

      readint.Item = "DeadKeys";
      if (Action(MT_ReadConfigInt, config, &readint) IS ERR_Okay) Self->DeadKeys = readint.Integer;

      acFree(config);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Reset
Short: Resets the keyboard settings.
###END###

*****************************************************************************/

static ERROR KEYBOARD_Reset(struct Keyboard *Self, APTR Void)
{
   Self->RepeatDelay = 250;
   Self->RepeatRate = 10.9;
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  SaveToObject
Short: Saves the current keyboard settings to another object.
###END###

*****************************************************************************/

static ERROR KEYBOARD_SaveToObject(struct Keyboard *Self, struct acSaveToObject *Args)
{
   struct mtWriteConfig write;
   OBJECTPTR config;
   UBYTE buffer[28];

   if (CreateObject(ID_CONFIG, NULL, &config, NULL,
      TAGEND) IS ERR_Okay) {

      write.Section = "KEYBOARD";
      write.Item = "Layout";
      write.Data = Self->prvLayout;
      Action(MT_WriteConfig, config, &write);

      write.Item = "RepeatDelay";
      write.Data = buffer;
      IntToStr(Self->RepeatDelay, buffer, sizeof(buffer));
      Action(MT_WriteConfig, config, &write);

      write.Item = "RepeatRate";
      write.Data = buffer;
      FloatToStr(Self->RepeatRate, buffer, sizeof(buffer));
      Action(MT_WriteConfig, config, &write);

      write.Item = "XShift";
      write.Data = buffer;
      IntToStr(Self->XShift, buffer, sizeof(buffer));
      Action(MT_WriteConfig, config, &write);

      ActionTags(AC_SaveToObject, config, Args->DestinationID, NULL);

      acFree(config);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Timer
*/

#ifdef __linux__

#include <sys/ioctl.h>

#include <sys/kd.h>

static ERROR KEYBOARD_Timer(struct Keyboard *Self, struct acTimer *Args)
{
   struct acDataChannel channel;
   struct dcKeyEntry entry;
   UBYTE buf[32];
   LONG i, nkeys, fd;
   UBYTE key;

   fd = GetResource(RES_KEYBOARDFD);
   while ((nkeys = read(fd, buf, sizeof(buf))) > 0) {
      for (i=0; i < nkeys; i++) {
         key = buf[i];

         entry.Flags = NULL;
         entry.Value = NULL;

         if (key & 0x80) {
            key = glKeyConversion[key & 0x7f];
            if (!key) key = K_SPACE;

            entry.Flags |= KQ_RELEASED;
            switch (key) {
               case K_LSHIFT:   glKeyFlags &= ~KQ_LSHIFT; break;
               case K_RSHIFT:   glKeyFlags &= ~KQ_RSHIFT; break;
               case K_LCONTROL: glKeyFlags &= ~KQ_LCONTROL; break;
               case K_RCONTROL: glKeyFlags &= ~KQ_RCONTROL; break;
               case K_LALT:     glKeyFlags &= ~KQ_LALT; break;
               case K_RALT:     glKeyFlags &= ~KQ_RALT; break;
               case K_LCOMMAND: glKeyFlags &= ~KQ_LCOMMAND; break;
               case K_RCOMMAND: glKeyFlags &= ~KQ_RCOMMAND; break;
            }
            entry.Flags |= glKeyFlags;
            entry.Value = key;
            glRepeatKey[key] = 0;
         }
         else {
            key = glKeyConversion[key];
            if (!key) key = K_SPACE;

            switch (key) {
               case K_LSHIFT:   glKeyFlags |= KQ_LSHIFT; break;
               case K_RSHIFT:   glKeyFlags |= KQ_RSHIFT; break;
               case K_LCONTROL: glKeyFlags |= KQ_LCONTROL; break;
               case K_RCONTROL: glKeyFlags |= KQ_RCONTROL; break;
               case K_LALT:     glKeyFlags |= KQ_LALT; break;
               case K_RALT:     glKeyFlags |= KQ_RALT; break;
               case K_LCOMMAND: glKeyFlags |= KQ_LCOMMAND; break;
               case K_RCOMMAND: glKeyFlags |= KQ_RCOMMAND; break;

               case K_CAPSLOCK:
                  SetResource(RES_PRIVILEGEDUSER, TRUE);
                  glKeyFlags ^= KQ_CAPSLOCK;
                  glLEDFlags ^= LED_CAP;
                  ioctl(GetResource(RES_CONSOLEFD), KDSETLED, glLEDFlags);
                  SetResource(RES_PRIVILEGEDUSER, FALSE);
                  break;

               case K_NUMLOCK:
                  SetResource(RES_PRIVILEGEDUSER, TRUE);
                  glKeyFlags ^= KQ_NUMLOCK;
                  glLEDFlags ^= LED_NUM;
                  ioctl(GetResource(RES_CONSOLEFD), KDSETLED, glLEDFlags);
                  SetResource(RES_PRIVILEGEDUSER, FALSE);
                  break;

               case K_SCROLL:
                  SetResource(RES_PRIVILEGEDUSER, TRUE);
                  glKeyFlags ^= KQ_SCROLLLOCK;
                  glLEDFlags ^= LED_SCR;
                  ioctl(GetResource(RES_CONSOLEFD), KDSETLED, glLEDFlags);
                  SetResource(RES_PRIVILEGEDUSER, FALSE);
                  break;
            }

            if (glRepeatKey[key]) entry.Flags |= KQ_REPEAT;
            glRepeatKey[key]++;

            entry.Flags |= glKeyFlags|KQ_PRESSED;
            entry.Value = key;
         }
#if 0
         if ((entry.Flags & KQ_ALT) AND (entry.Flags & KQ_CTRL)) {
            if ((entry.Value >= K_F1) AND (entry.Value <= K_F9)) {
               /*** Holding Ctrl-Alt-F1/9 switches terminals ***/
               if (ioctl((GetResource(RES_CONSOLEFD), VT_ACTIVATE, entry.Value - K_F1 + 1) IS -1) {
                  DPrintF("@Keyboard:","Failed to switch to terminal %d.", entry.Value - K_F1 + 1);
               }
               else DPrintF("Keyboard:","Switched to terminal %d.", entry.Value - K_F1 + 1);
            }
         }
#endif
         channel.ObjectID     = NULL;
         channel.DataType     = DATA_KEYPRESS;
         channel.Version      = 1;
         channel.Buffer       = &entry;
         channel.Size         = sizeof(struct dcKeyEntry);
         channel.TotalEntries = 1;
         Action(AC_DataChannel, Self, &channel);
      }
   }

   return(ERR_Okay);
}
#endif

/*****************************************************************************

###METHOD###
Name:    TranslateChars()
Short:   Translates printable characters into their equivalent key codes.
Version: 2.0

###DESCRIPTION###
<p>The TranslateChars method is used to convert UTF-8 character strings
into their equivalent raw key values.  The routine will scan the character
buffer in the String argument and write the equivalent key values to the
buffer referred to by the Keys argument.  The routine will stop when the
String buffer is terminated, or the Keys buffer runs out of space.</p>

###INPUT###
STRING/String: String of character codes to be translated.
APTR/Keys: Output buffer (struct dcKeyEntry).
LONG/Size: Size of the output buffer in bytes.

###RESULT###
ERR_Okay: The characters were successfully translated.
ERR_Args: Invalid arguments were detected.

###END###

*****************************************************************************/

static ERROR KEYBOARD_TranslateChars(struct Keyboard *Self, struct mtTranslateChars *Args)
{
   if ((!Args) OR (!Args->String) OR (!Args->Keys) OR (Args->Size < sizeof(struct dcKeyEntry))) {
      return(ObjectError(Self, MT_TranslateChars, ERR_Args));
   }





   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    TranslateKeys()
Short:   Translates key codes into printable character values
Version: 2.0

###DESCRIPTION###
<p>The TranslateKeys method is used to convert raw key values into printable
characters.  The routine will scan the key array passed in the Keys argument
and each key value that is linked to a printable character will be written
to the String buffer that you have supplied.  The routine will stop when the
the Keys array is terminated, or the supplied buffer runs out of space.</p>

<p>Please note that this method will return unicode (UTF-8) characters for
keypresses that translate to unicode values.  For this reason the minimum
recommended buffer size is 7 bytes.  It is also worth keeping in mind that
even a single keypress can potentially return multiple characters if a string
sequence has been assigned to the key.</p>

###INPUT###
APTR/Keys: List of keys to translate (struct dcKeyEntry), terminated with a NULL entry.
STRING/String: String based output buffer.
LONG/Size: Size of the output buffer in bytes.

###RESULT###
ERR_Okay: The characters were successfully translated.
ERR_Args: Invalid arguments were detected.
ERR_EmptyString: None of the keys were translated into printable characters.

###END###

*****************************************************************************/

static ERROR KEYBOARD_TranslateKeys(struct Keyboard *Self, struct mtTranslateKeys *Args)
{
   struct dcKeyEntry *array;
   LONG pos;
   UWORD value;

   if ((!Args) OR (!Args->Keys) OR (!Args->String) OR (Args->Size < 1)) {
      return(ObjectError(Self, MT_TranslateKeys, ERR_Args));
   }

   pos = 0;
   for (array=Args->Keys; (array->Value) AND (pos < Args->Size-1); array++) {
      if (array->Value < K_LIST_END) {
         if (array->Flags & KQ_CONTROL) continue;

         if (Self->XShift IS XSF_ALTGR) {
            if (array->Flags & (KQ_LALT|KQ_COMMAND)) continue;
         }
         else if (array->Flags & KQ_ALT) continue;

         if ((Self->XShift IS XSF_ALTGR) AND (array->Flags & KQ_RALT)) {
            if (array->Flags & KQ_SHIFT) value = Self->prvKeymap[array->Value][3];
            else value = Self->prvKeymap[array->Value][2];
         }
         else if ((Self->XShift IS XSF_COMMAND) AND (array->Flags & KQ_COMMAND)) {
            if (array->Flags & KQ_SHIFT) value = Self->prvKeymap[array->Value][3];
            else value = Self->prvKeymap[array->Value][2];
         }
         else if (array->Flags & KQ_SHIFT) {
            value = Self->prvKeymap[array->Value][1];
         }
         else value = Self->prvKeymap[array->Value][0];

         if (value) {
            if (array->Flags & KQ_DEADKEY) {
               /*** Deadkey support ***/
               value = convertdeadkey(Self->prvDeadKey, value);
               if (!value) continue;
            }

            /*** Standard printable character ***/
            if ((array->Flags & KQ_CAPSLOCK) AND (value >= 'a') AND (value <= 'z')) value = (value - 'a') + 'A';
            pos += WriteUTF8(Args->String+pos, value);
         }
      }
   }

   Args->String[pos] = 0;

   if (pos < 1) return(ERR_EmptyString); /* No keys were translated */

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   RepeatDelay
Short:  Defines the delay between the holding of any key and the start of the key repeat rate.
Type:   LONG
Status: Get/Set

###DESCRIPTION###
<p>Defines the delay, in milliseconds, between the time of pressing a key and
the start of the key repeat rate.  The recommended default is 250ms.</p>

###END###

*****************************************************************************/

static ERROR GET_RepeatDelay(struct Keyboard *Self, LONG *Value)
{
   *Value = Self->RepeatDelay;
   return(ERR_Okay);
}

static ERROR SET_RepeatDelay(struct Keyboard *Self, LONG *Value)
{
   if (*Value <= 0) Self->RepeatDelay = *Value;
   else if (*Value > 1000) Self->RepeatDelay = 1000;
   else Self->RepeatDelay = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Layout
Short:  Determines the keyboard layout (defaults to US101).
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The layout of a keyboard can be redefined by setting the Layout field.
The available keyboard layouts are determined by the available configuration
files in the &quot;system:config/keymaps/&quot; directory.  For instance,
if an &quot;italian&quot; file is available, you can set a layout of
&quot;Italian&quot;.</p>

<p>If you set a layout string that does not match any existing keymap files,
the keyboard object will revert to the default US101 format.</p>

###END###

*****************************************************************************/

static ERROR GET_Layout(struct Keyboard *Self, STRING *Value)
{
   *Value = Self->prvLayout;
   return(ERR_Okay);
}

static ERROR SET_Layout(struct Keyboard *Self, STRING Value)
{
   LONG i;

   if ((Value) AND (*Value)) {
      for (i=0; (i < LEN_LAYOUT-1) AND (Value[i]); i++) {
         Self->prvLayout[i] = Value[i];
      }
      Self->prvLayout[i] = 0;
   }

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      LoadKeymap(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   RepeatRate
Short:  Defines the keyboard repeat rate for held keypresses.
Type:   LONG
Status: Get/Set

###DESCRIPTION###
<p>Defines the character repeat rate in characters per second.</p>

###END###

*****************************************************************************/

static ERROR GET_RepeatRate(struct Keyboard *Self, FLOAT *Value)
{
   *Value = Self->RepeatRate;
   return(ERR_Okay);
}

static ERROR SET_RepeatRate(struct Keyboard *Self, FLOAT *Value)
{
   if (*Value <= 0) Self->RepeatRate = *Value;
   else if (*Value > 50) Self->RepeatRate = 50;
   else Self->RepeatRate = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   UnicodeTable
Short:  Refers to a global unicode table for converting unicode values to names and vice versa.
Type:   struct UnicodeDescriptor *
Status: Get

###DESCRIPTION###
<p>This field refers to a unicode table that you can use to translate unicode
values into their standard system names.  The structure of
each entry in the table is defined as follows:</p>

<pre>
   struct UnicodeDescriptor {
      STRING Name;
      UWORD  Value;
   };
</pre>

<p>The table is sorted by unicode value (ascending) and is terminated with a
NULL entry in both fields.  You can view all of the content of the table
if you load the &quot;input/unicodetable.h&quot; include file.</p>

###END###

*****************************************************************************/

static ERROR GET_UnicodeTable(struct Keyboard *Self, struct UnicodeDescriptor **Table)
{
   *Table = glUnicodeTable;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XShift
Short:  Configures the extended-shift key.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The extended-shift key is normally mapped against the AltGr key for foreign
keyboards; also known as the right-alt key for vanilla keyboards.  You can change
the extended shift key to the left and right command keys if you pass XSF_COMMAND
to the XShift field.  Switching to the default AltGr key is possible by passing
XSF_ALTGR.</p>

<p>The extended-shift key has the same properties as the keyboard's standard
shift key, i.e. the user holds down the key and presses it in conjunction with
another key to create a printable character.  The resulting character will be taken from
the keyboard's mapping for the second shift key.  If there is no character set
against the extended-shift key, no printable character will result.</p>

###END###

*****************************************************************************/

static ERROR SET_XShift(struct Keyboard *Self, LONG *Value)
{
   if ((*Value IS XSF_ALTGR) OR (*Value IS XSF_COMMAND)) {
      Self->XShift = *Value;
      return(ERR_Okay);
   }
   else {
      DPrintF("@SetField:","[Keyboard] Invalid XShift value of %d.", *Value);
      return(ERR_OutOfRange);
   }
}

/*****************************************************************************
** Internal: LoadKeymap()
*/

static ERROR LoadKeymap(struct Keyboard *Self)
{
   struct acRead read;
   UBYTE buffer[80], *byteswitch, tmpchar;
   UWORD *keytable;
   OBJECTPTR file;
   LONG i, j, pos, type, headlength, maxkeys, totalkeys, buffersize, keyvalue;

   pos = StrCopy("system:config/keymaps/", buffer, sizeof(buffer));
   StrCopy(Self->prvLayout, buffer+pos, sizeof(buffer)-pos);

   if (CreateObject(ID_FILE, NF_CHILD, (OBJECTPTR *)&file, NULL,
      FID_Location|TSTRING, buffer,
      FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
      TAGEND) IS ERR_Okay) {

      read.Buffer = buffer;
      read.Length = 8;
      if ((Action(AC_Read, file, &read) IS ERR_Okay) AND (read.Length IS 8)) {
         if (StrCompare("KEYMAP10", buffer, 8, STR_MATCHCASE) IS ERR_Okay) {
            ReadLong(file, &headlength);
            ReadLong(file, &totalkeys);
            ReadLong(file, &maxkeys);

            ActionTags(AC_Seek, file, (DOUBLE)headlength, SEEK_START);

            /*** Read the key values ***/

            buffersize = totalkeys * maxkeys * sizeof(UWORD);
            if (AllocMemory(buffersize, MEM_DATA|MEM_NOCLEAR, &keytable, NULL) IS ERR_Okay) {
               read.Buffer = (APTR)keytable;
               read.Length = buffersize;
               Action(AC_Read, file, &read);

               #ifdef REVERSE_BYTEORDER
                  byteswitch = (UBYTE *)keytable;
                  for (i=0; i < buffersize-1; i+=2) {
                     tmpchar = byteswitch[i];
                     byteswitch[i] = byteswitch[i+1];
                     byteswitch[i+1] = tmpchar;
                  }
               #endif

               j = 0;
               for (i=0; i < totalkeys; i++) {
                  for (type=0; (type < 4) AND (type < maxkeys); type++) {
                     if ((keyvalue = keytable[j]) < K_LIST_END) {
                        Self->prvKeymap[keyvalue][type] = keytable[j+1+type];
                     }
                  }
                  j += maxkeys;
               }

               FreeMemory(keytable);
            }
         }
      }
      else {
         acFree(file);
         DPrintF("!Keyboard:","Keymap file \"%s\" is empty.", buffer);
         return(ERR_GetField);
      }

      acFree(file);
   }
   else {
      DPrintF("!Keyboard:","Failed to load keymap \"%s\".", buffer);
      return(ERR_CreateObject);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: isdeadkey()
*/

static LONG isdeadkey(UWORD Unicode)
{
   WORD i;

   if ((Unicode < 0x2b0) OR (Unicode > 0x36f)) return(FALSE);

   for (i=0; i < ARRAYSIZE(deadkeys); i++) {
      if (Unicode IS deadkeys[i].deadkey) return(TRUE);
   }

   return(FALSE);
}

/*****************************************************************************
** Internal: convertdeadkey()
**
** NB: A better way to support deadkeys would be to simply output the unicode
** character and follow it with the combining unicode character.  The font
** system just needs to have the smarts to combine the two together when drawing
** them.
*/

static UWORD convertdeadkey(UWORD DeadKey, UWORD Unicode)
{
   struct deadchar *table;
   WORD i;

   for (i=0; i < ARRAYSIZE(deadkeys); i++) {
      if (DeadKey IS deadkeys[i].deadkey) {
         table = deadkeys[i].convert;
         while (table->unicode) {
            if (table->unicode IS Unicode) return(table->conversion);
            table++;
         }
      }
   }
   return(Unicode);
}
